import java.util.*;

public class SymbolTable {

	public final static int CLASS_VAR  = 0;
	public final static int CLASS_FUNC = 1;
	public final static int CLASS_PARAM = 2;
	
	private Vector hashVector = new Vector();
	private VarAddressVector varAddressVector = new VarAddressVector();
	private SymbolTableEntry symtabEntryOfCurrentFunction = null;
	private int currentLevel = 0;

	public SymbolTableEntry put(String key, int type, int idclass) {
		assert hashVector.size()>=1: "Hashvector enthaelt keine Elemente (Hashtabellen)";

		Hashtable table = (Hashtable) hashVector.lastElement();
		
		if (table.get(key) != null) Minako.error("Identifier already defined -- " + key);

		int address;
		if (currentLevel > 1) address = varAddressVector.getNextAddress();
		else address = 0;

		SymbolTableEntry symtabEntry = new SymbolTableEntry(key, address, type, currentLevel, idclass);
		
		table.put(key, symtabEntry);
		if (idclass == CLASS_FUNC) symtabEntryOfCurrentFunction = symtabEntry;
		
		return symtabEntry;
	}

	public void enterBlock() {
		currentLevel++;
		varAddressVector.enterBlock();
		Hashtable hashtable = new Hashtable(10);
		hashVector.addElement(hashtable);
		assert(currentLevel==hashVector.size());
	}

	public void leaveBlock() {
		varAddressVector.leaveBlock();
		currentLevel--;
		hashVector.setSize(currentLevel);
	}

	public SymbolTableEntry getSymbolTableEntry(String key) {
		Object o = null;
		for (int i = hashVector.size()-1; (o == null) && (i>=0); i--) {
			Hashtable table = (Hashtable) hashVector.elementAt(i);
			o = table.get(key);
		}
		if (o == null) Minako.error(key + " ist nicht in Symboltabelle");
		return (SymbolTableEntry) o;
	}

	public SymbolTableEntry getSymbolTableEntryOfCurrentFunction() {
		return symtabEntryOfCurrentFunction;
	}

	public int getCurrentLevel() {
		return currentLevel;
	}

	public String toString() {
		StringBuffer buffer = new StringBuffer();
		Enumeration e = hashVector.elements();
		while (e.hasMoreElements()) {
			Hashtable hashtable = (Hashtable) e.nextElement();
			Enumeration f = hashtable.keys();
			while (f.hasMoreElements()) {
				Object key = f.nextElement();
				buffer.append(key + " " + hashtable.get(key) + '\n');
			}
		}
		return buffer.toString();
	}
	
	// ------------------------------------------------------------

	class VarAddressVector {
		private Stack stack = new Stack();

		public int getNextAddress() {
			int address = ((Integer) stack.pop()).intValue();
			stack.push(new Integer(address+1));
			return address;
		}

		public void enterBlock() {
			if (currentLevel==1) stack.push(new Integer(0));
			else stack.push(new Integer(((Integer) stack.lastElement()).intValue()));
		}

		public void leaveBlock() {
			stack.pop();
		}

		private void printAddressStack() {
			Enumeration e = stack.elements();
			for (int i=0; e.hasMoreElements(); i++) {
				System.out.println("Level " + i + " -- " + e.nextElement());
			}
		}
	}

}
